import argparse
import sys
from zope import component
from zope import event
from zope import interface
from zope.configuration.xmlconfig import XMLConfig
from zope.interface.exceptions import DoesNotImplement
from sparc.configuration import zcml
from sparc.configuration.container.zcml import IZCMLFiles
from sparc.configuration.events import SparcApplicationConfiguredEvent
from sparc.logging import logging

import zope.component.event #needed in order to initialize the event notification environment

from . import ISparcPyContainerConfiguredApplication


def getScriptArgumentParser(description, args=sys.argv):
    """Return ArgumentParser object
    
    Args:
        description: Text description of application ArgumentParser will be
                     applied to.
    
    Kwargs:
        args (list):        list of arguments that will be parsed.  The default
                            is the sys.argv list, and should be correct for most
                            use cases.
    
    Returns:
        ArgumentParser object that can be used to validate and execute the
        current script invocation.
    """
    # Description
    parser = argparse.ArgumentParser(
            description=description)

    # config_file
    parser.add_argument('config_file',
            help="Valid script configuration file.  This should be the path to "\
                 "the script YAML configuration file.  See config_sample.yaml"\
                 "for detailed specifications.")
    
    # --verbose
    parser.add_argument('--verbose',
            action='store_true',
            help="Echo verbose messages to stdout.")
    
    # --debug
    parser.add_argument('--debug',
            action='store_true',
            help="Echo debug messages to stdout.")
    
    return parser


@interface.implementer(ISparcPyContainerConfiguredApplication)
class YamlCliAppMixin(object):
    def __init__(self, config=None, verbose=False, debug=False):
        """Init
        
        Args:
            config: ISparcAppPyContainerConfiguration compatible Py container for runtime configuration (i.e. a dict or list)
            verbose: True indicates verbose logging
            debug: True indicate debug logging
        """
        self.setLoggers(verbose, debug)
        #Setup the Zope component registry
        zcml.Configure(packages=[self.app_zcml]) # base app component registrations
        self.logger.debug("Zope Component Registry initialized")
        if not config:
            config = {}
        self._config = component.createObject(u'sparc.configuration.container', config) # initialize
    
    def get_config(self):
        return self._config
    
    def set_config(self, config):
        if not ISparcPyContainerConfiguredApplication.providedBy(config):
            config = component.createObject(u'sparc.configuration.container', config)
        self._config = config

    def configure(self):
        #Search/load any additional runtime config zcml
        for z_file in IZCMLFiles(self.get_config()):
            XMLConfig(z_file.file, z_file.package)()
            self.logger.info("zcml configuration processed for {}:{}".format(z_file.package.__name__ if z_file.package else None, z_file.file))
        event.notify(SparcApplicationConfiguredEvent(self))

    def setLoggers(self, verbose, debug):
        logger = logging.getLogger() # root logger
        if verbose:
            logger.setLevel('INFO')
        if debug:
            logger.setLevel('DEBUG')

    # OVERRIDES
    #  Class extenders need to implement these artifacts
    
    logger = None #override with app specific logger
    app_zcml = None #Optional, override with tuple (py_package_object, 'configure.zcml')
    def go(self):
        raise DoesNotImplement()


from zope import interface

class ISparcYamlDocuments(interface.Interface):
    def documents(config):
        """Generator of 
        sparc.configuration.container.ISparcAppPyContainerConfiguration objects
        based on config
        
        All Python container based items within the top level 
        ISparcAppPyContainerConfiguration will also be marked as
        ISparcAppPyContainerConfiguration objects.
        
        Args:
            config: Unicode valid file path to a Yaml configuration or a valid
                    Yaml content string.
        """
    def first(config):
        """sparc.configuration.container.ISparcAppPyContainerConfiguration 
        object for first document in config
        
        Args:
            config: [same as documents()]
        """


from zope import interface
from zope.interface.interfaces import ObjectEvent
from .interfaces import ISparcApplicationConfiguredEvent

@interface.implementer(ISparcApplicationConfiguredEvent)
class SparcApplicationConfiguredEvent(ObjectEvent):
    """A configuration has been applied to a ISparcApplication"""


from zope import interface
from .. import ISparcAppConfiguration
from .. import ISparcApplication

class ISparcAppPyContainerConfiguration(ISparcAppConfiguration):
    """Marker for a dict/list/tuple/set based Sparc app config
    
    This type of configuration will be based on a native Python data container.
    """

class ISparcPyContainerConfiguredApplication(ISparcApplication):
    def get_config():
        """Return ISparcAppPyContainerConfiguration provider for current 
           runtime configuration settings"""
    def set_config(config):
        """Set ISparcAppPyContainerConfiguration provider or Python container 
           (dict, list, tuple, set) for current runtime configuration settings
        
        Args:
            config: ISparcAppConfiguration provider
        """

class ISparcPyDictValueIterator(interface.Interface):
    """Find Yaml document values based on a key"""
    def values(config, key):
        """Iterator of values in document matching given key
        
        This acts based on the top level document type.  For dict, this will 
        search dict keys and return the corresponding value.
        
        For others, this will iterate the container and search for dicts.
        Each dict found will be processed as above.
        
        Note:
            This is not a recursive method, it will only search 1 level deep
            on non-dict  
        
        Args:
            config: ISparcAppPyContainerConfiguration provider.
            key: Hashable key to match on
        """

class IPyContainerConfigValue(interface.Interface):
    """Find Yaml document values based on a key"""
    def get(*args, **kwargs):
        """Get the first config value matching key
        
        See ISparcPyDictValueIterator for information on how search for key
        is performed.
        
        Raises: KeyError if key is not found and default is not given
        
        Args:
            Ordered keys to search in config.  To find 'value1' in this
            container {'key1': {'key2': 'value1'}}, you would make a call
            to get('key1', 'key2')
        Kwargs:
            default: value to return if given args are not referencable.
            
        Returns: object value from key or default
        """
    def query(*args):
        """Get the first config value matching key, if available
        
        See ISparcPyDictValueIterator for information on how search for key
        is performed.
        
        Args:
            [see get()]
            
        Returns: object value from key or None if key is not found
        """



from zope import interface

class IZCMLFile(interface.Interface):
    """ZCML File descriptor"""
    file = interface.Attribute("Sting zcml file name")
    package = interface.Attribute("Python module object of package containing file")

class IZCMLFiles(interface.Interface):
    """Identifies ZCML files"""
    def __iter__():
        """Iterator of IZCMLFile providers"""


from xml.etree import ElementTree
from zope import interface
from zope import component
from zope.component import IFactory
from sparc import proxy
from .interfaces import IAppElementTreeConfig

@interface.implementer(IFactory)
class AppConfigFactory(object):
    
    title = u"Create an XML based app configuration"
    description = u"returns a xml.etree.ElementTree marked with IAppElementTreeConfig"
    
    def __call__(self, xml_config):
        config = ElementTree.parse(xml_config).getroot()
        try:
            interface.alsoProvides(config, IAppElementTreeConfig)
        except AttributeError: #Py3 breaks
            config = component.getUtility(proxy.IZopeInterfaceProviderProxy)(config)
            interface.alsoProvides(config, IAppElementTreeConfig)
        return config
    
    def getInterfaces(self):
        etree = ElementTree.ElementTree()
        interface.alsoProvides(etree, IAppElementTreeConfig)
        return interface.providedBy(etree)



from zope import interface
from zope.component import getSiteManager
from zope.component.interfaces import ComponentLookupError
from zope.component import getUtility
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.interfaces import IVocabularyRegistry
from zope.schema.vocabulary import setVocabularyRegistry
import zope.configuration.xmlconfig
import zope.component
import zope.security

from sparc.logging import logging
logger = logging.getLogger(__name__)

#Copied from Zope2.App.schema
@interface.implementer(IVocabularyRegistry)
class Zope2VocabularyRegistry(object):
    """IVocabularyRegistry that supports global and local utilities.
    """

    __slots__ = ()

    def get(self, context, name):
        """See zope.schema.interfaces.IVocabularyRegistry.
        """
        factory = getUtility(IVocabularyFactory, name)
        return factory(context)

#Copied from Zope2.App.schema
def configure_vocabulary_registry():
    setVocabularyRegistry(Zope2VocabularyRegistry())


def Configure(packages = None):
    """Setup Zope Component global registry
    
    Usage:
        Simple configuration of ZCA directives, makes ZCA available for use
        >>> Configure()
        
        Configure a package that has a configure.zcml file available
        >>> import sparc.cache
        >>> Configure([sparc.cache])
        
        Configure a list of packages, some of which require a custom zcml spec
        >>> import sparc.cache, sparc.db
        >>> myPackages = [
        ...               sparc.cache, # loads the configure.zcml file in package
        ...               sparc.db,  # loads the configure.zcml file in package
        ...               (sparc.db, 'memory.zcml') # loads the memory.zcml file in package
        ...              ]
    
    Args:
        packages: A package or list of packages that are to be configured.
                  By default, each package is expected to have a configure.zcml
                  file located in the package root.  Optionally, a package
                  can be identified as a sequence whose first entry is the
                  Python package, and whose second entry is the zcml file name
                  (string) that is to be configured.
    """
    try:
        iter(packages)
        packages = list(packages)
    except TypeError:
        packages = [packages]
    packages = [
                 (zope.component, 'configure.zcml'),
                 (zope.security, 'configure.zcml')
                 ] + packages
    for config in packages:
        try:
            iter(config)
        except:
            config = (config, 'configure.zcml',)
        package, zcml = config
        zope.configuration.xmlconfig.XMLConfig(zcml, package)()
        logger.debug("Configured package %s with zcml file %s", str(package), str(zcml))
    #This allows vocabulary lookups from schemas via IVocabularyFactory interface
    configure_vocabulary_registry()


def ConfigurationRequired(classThatRequiresConfigration):
    """Decorator for callables that require ZCA registry configuration
    
    Mostly, the sparc framework does not require use of ZCA.  The *.zcml
    files allows users of the framework to interact via ZCA, however, it
    is not a strict requirement for use.  Because of this, most things in 
    sparc will not utilize component lookups.
    
    However, on occasion, it makes sense to put common code such as mixins
    that *do* require component lookups.  For these cases, we decorate them 
    in order to make it clear that these items require a configured registry
    to function.
    """
    def _checkForZCARegistry(*args, **kwargs):
        sm = getSiteManager()
        configured = False
        for a in sm.registeredAdapters():
            configured = True
            break
        for h in sm.registeredSubscriptionAdapters():
            configured = True
            break
        for u in sm.registeredUtilities():
            configured = True
            break
        for h in sm.registeredHandlers():
            configured = True
            break
        if not configured:
            raise ComponentLookupError("Use of this class requires sparc component registration.")
        return classThatRequiresConfigration(*args, **kwargs)
    return _checkForZCARegistry


from zope import component
from zope import interface
from sparc.configuration import container

@interface.implementer(container.ISparcPyDictValueIterator)
class SparcPyContainerValueIterator(object):
    
    def values(self, config, key):
        _list = config if not isinstance(config, dict) else [config]
        for i in _list:
            if not isinstance(i, dict):
                continue
            for k, v in i.items():
                if k == key:
                    yield v

@interface.implementer(container.IPyContainerConfigValue)
@component.adapter(container.ISparcAppPyContainerConfiguration)
class PyContainerConfigValue(object):
    def __init__(self, context):
        self.context = context
    
    #def get(self, key1, key2, ..., default=None): #<--signature
    def get(self, *args, **kwargs):
        has_default = True if 'default' in kwargs else False
        
        config = self.context
        for key in args:
            values = list(SparcPyContainerValueIterator().values(config, key))
            if not values:
                if has_default:
                    return kwargs['default']
                else:
                    raise KeyError(
                        "key: {} not available in configuration.".format(key))
            config = values[0] # set for next loop, if there is one
            v = values[0]
        return v

    def query(self, key):
        try:
            return self.get(key)
        except KeyError:
            return None


import importlib
from zope import component
from zope.component.factory import Factory
from zope import interface

from sparc.configuration import container
from . import IZCMLFile
from . import IZCMLFiles

@interface.implementer(IZCMLFile)
class ZCMLFile(object):
    def __init__(self, package=None, file_=None):
        self.package = package
        self.file = file_
ZCMLFileFactory = Factory(ZCMLFile)

@interface.implementer(IZCMLFiles)
@component.adapter(container.ISparcAppPyContainerConfiguration)
class ZCMLFilesFromContainerConfig(object):
    def __init__(self, context):
        self.context = context
    
    def __iter__(self):
        
        zcml_values = list(component.getUtility(container.\
                        ISparcPyDictValueIterator).values(self.context, 
                                                          'ZCMLConfiguration'))
        
        for zcml_entry in zcml_values:
            zcml_entry = zcml_entry if not isinstance(zcml_entry, dict) else [zcml_entry]
            for zcml_config in zcml_entry:
                file_name = zcml_config.get('file', 'configure.zcml')
                yield ZCMLFile(
                    package = importlib.import_module(zcml_config['package']) \
                                            if 'package' in zcml_config else None,
                    file_ = file_name)


from zope import interface
from zope.interface.interfaces import IObjectEvent

class ISparcApplicationConfiguredEvent(IObjectEvent):
    """A configuration has been applied to a ISparcApplication"""

class ISparcAppConfiguration(interface.Interface):
    """Marker for Sparc application configuration data"""

class ISparcApplication(interface.Interface):
    """An Application"""
    def get_config():
        """Return ISparcAppConfiguration provider for current runtime configuration settings"""
    def set_config(config):
        """Set ISparcAppConfiguration for current runtime configuration settings
        
        Args:
            config: ISparcAppConfiguration provider
        """
    def configure():
        """Configure runtime environment to settings to current config.
        Issues a ISparcApplicationConfiguredEvent
        """
    def go():
        """Run the application"""


from zope import interface
from zope.component.factory import Factory
from . import ISparcAppPyContainerConfiguration

@interface.implementer(ISparcAppPyContainerConfiguration)
class SparcAppPyContainerConfiguration(object):

    @classmethod
    def is_container(cls, value):
        return True if (isinstance(value, dict)  or \
                                isinstance(value, list) or \
                                isinstance(value, tuple) or \
                                isinstance(value, set)) \
                                else False

    @classmethod
    def create(cls, container):
        """Return shallow-copied container marked with ISparcAppPyContainerConfiguration"""
        if isinstance(container, dict):
            return SparcAppPyDictConfiguration(container)
        if isinstance(container, list):
            return SparcAppPyListConfiguration(container)
        if isinstance(container, tuple):
            return SparcAppPyTupleConfiguration(container)
        if isinstance(container, set):
            return SparcAppPySetConfiguration(container)
        raise ValueError('Expected container to be dict, list, tuple, or set. Got: {}'.format(type(container)))

    def __new__(cls, container):
        config = cls.create(container)
        if isinstance(config, dict):
            for k,v in config.items():
                if cls.is_container(v):
                    config[k] = cls(v)
        if isinstance(config, list):
            for i,v in enumerate(config):
                if cls.is_container(v):
                    config[i] = cls(v)
        if isinstance(config, set):
            for v in config:
                if cls.is_container(v):
                    config.delete(v)
                    config.add(cls(v))
        if isinstance(config, tuple):
            new_tuple = [] # we'll convert to tuple below...need it mutable for now
            has_container = False
            for v in config:
                if cls.is_container(v):
                    has_container = True
                    new_tuple.append(cls(v))
                else:
                    new_tuple.append(v)
            if has_container:
                config = tuple(new_tuple)
        return config
sparcAppPyContainerConfigurationFactory = Factory(SparcAppPyContainerConfiguration)

@interface.implementer(ISparcAppPyContainerConfiguration)
class SparcAppPyDictConfiguration(dict):
    pass
sparcAppPyDictConfigurationFactory = Factory(SparcAppPyDictConfiguration)

@interface.implementer(ISparcAppPyContainerConfiguration)
class SparcAppPyListConfiguration(list):
    pass
sparcAppPyListConfigurationFactory = Factory(SparcAppPyListConfiguration)

@interface.implementer(ISparcAppPyContainerConfiguration)
class SparcAppPyTupleConfiguration(tuple):
    pass
sparcAppPyTupleConfigurationFactory = Factory(SparcAppPyTupleConfiguration)

@interface.implementer(ISparcAppPyContainerConfiguration)
class SparcAppPySetConfiguration(set):
    pass
sparcAppPySetConfigurationFactory = Factory(SparcAppPySetConfiguration)



from os.path import isfile
import yaml
from zope import interface

from sparc.configuration.container import container
from . import ISparcYamlDocuments

@interface.implementer(ISparcYamlDocuments)
class SparcYamlDocuments(object):

    def documents(self, config):
        config = config if not isfile(config) else open(config)
        for doc in yaml.load_all(config):
            if container.SparcAppPyContainerConfiguration.is_container(doc):
                yield container.sparcAppPyContainerConfigurationFactory(doc)
            else:
                yield doc

    def first(self, config):
        return next(self.documents(config))


from zope.interface import Interface

class IElementTree(Interface):
    """Marker for an ElementTree instance"""

class IXMLElement(Interface):
    """Marker for an ElementTree Element instance"""

class IAppElementTreeConfig(IElementTree):
    """Marker for a ElementTree based application configuration instance"""


from .interfaces import IPyContainerConfigValue
from .interfaces import ISparcAppPyContainerConfiguration
from .interfaces import ISparcPyContainerConfiguredApplication
from .interfaces import ISparcPyDictValueIterator